#!/usr/bin/env python3
import json
import pprint

PORTS_OF_INTEREST = {80, 443}


async def parse_internetdb(args):
    """program made to be able to parse shodan internet db output,
    which looks like this:
    {
     "cpes":        [],
     "hostnames":   ["dns.google"],
     "ip":          "8.8.8.8",
     "ports":       [53, 443],
     "tags":        [],
     "vulns":       []
    },

    """

    with open(args.input_file, "r") as f:
        for line in f.readlines():
            shodan = json.loads(line)

            if shodan.get("vulns"):
                pprint.pprint(shodan)

            if ip := shodan.get("ip"):
                if ip == "YOUR IP":
                    pprint.pprint(shodan)

            elif ports := shodan.get("ports"):
                if PORTS_OF_INTEREST & set(ports):
                    pprint.pprint(shodan)



"""
__author__ = "pimvh"
__copyright__ = "No Copyright 2022"
__credits__ = ["John Matherly", "Shodan"]
__license__ = "GPL"
__version__ = "1.0.1"
__maintainer__ = "Pim van Helvoirt"
__email__ = "pim.van.helvoirt@home.nl"
__status__ = "Production"
"""
import argparse
import asyncio
import json
import logging
import os
from dataclasses import asdict
from io import TextIOWrapper
from ipaddress import ip_network

import aiohttp

from .models import ShodanResult

logger = logging.getLogger()
logging.basicConfig(level=os.getenv("LOGLEVEL", logging.INFO))

SHODAN_URL = "https://internetdb.shodan.io/"


class ShodanLookupper:
    """class that looks up IPs from shodan by opening the passed file, if it exist,
    and starting a HTTP session against Shodan's internet db.
    Retrieve all information"""

    def __init__(
        self, ip_addrs: asyncio.Queue, output_file, output_type, skipmissing
    ) -> None:
        """initialized passed params to attributes"""

        self.shodan_url = SHODAN_URL
        self.ips: asyncio.Queue = ip_addrs
        self.output_file = output_file
        self.output_type = output_type
        self.skipmissing = skipmissing
        self._logger = logging.getLogger("ShodanLookupper")

    async def look_up_ips(self):
        """open passed file if present, start a session and close the file, if present"""

        if self.output_file:
            with open(self.output_file, "w+") as f:
                await self.start_session(f)
        else:
            await self.start_session(None)

    async def output_str(self, output_file: TextIOWrapper, resp: ShodanResult):
        """helper to output str output of Shodan"""

        print(str(resp), file=output_file)
        print("\n\n", file=output_file)

    async def output_json(self, output_file: TextIOWrapper, resp: ShodanResult):
        """helper to output JSON output of Shodan"""

        print(json.dumps(asdict(resp)), file=output_file)

    async def start_session(self, output_file):
        """start a Session against Shodan's InternetDB,
        and retrieve information of IPs until the queue is empty"""

        self._logger.info("Starting session...")

        async with aiohttp.ClientSession() as session:
            while not self.ips.empty():
                try:
                    ip = self.ips.get_nowait()
                    resp = await session.get(
                        self.shodan_url + ip, headers={"User-Agent": "nrich"}
                    )

                    resp = ShodanResult(**await resp.json())

                    self._logger.info("Got response: %s...", resp)

                    if self.skipmissing:
                        if resp.empty:
                            continue

                    if self.output_type == "str":
                        await self.output_str(output_file, resp)
                    else:
                        await self.output_json(output_file, resp)

                except json.JSONDecodeError:
                    logger.warning("Value error reading from JSON from Shodan request.")
                    continue
                except (
                    aiohttp.ClientConnectionError,
                    aiohttp.ClientResponseError,
                ) as exc:
                    logger.warning("aioHTTP error: %s", exc)
                    continue
                except (KeyError, ValueError) as exc:
                    logger.warning("Value error reading from Shodan request: %s", exc)
                    continue
                except asyncio.QueueEmpty as exc:
                    logger.warning("AsyncIO Queue Error: %s", exc)
                    break

                finally:
                    self.ips.task_done()


async def ip_reader(ip_file, queue: asyncio.Queue):
    """read IPs from a file and put them on a Queue"""

    with ip_file as f:
        ips: list = f.readlines()

        for ip in ips:
            try:
                network = ip_network(ip.strip())

                for ip in network:
                    queue.put_nowait(str(ip))

            except ValueError:
                continue


async def nrich(args: argparse.Namespace):
    logger.info("creating queue and IPlookup object...")

    queue = asyncio.Queue()
    ip_lookup = ShodanLookupper(
        queue, args.output_file, args.output_type, args.skip_missing
    )

    logger.info("creating tasks...")

    input_task = asyncio.create_task(
        ip_reader(args.input_file, queue), name="IP inputter"
    )
    lookup_task = asyncio.create_task(ip_lookup.look_up_ips(), name="IP lookupper")

    tasks = [input_task, lookup_task]

    # sleep shortly to avoid directly closing an empty queue.
    await asyncio.sleep(0.1)

    logger.info("running tasks...")
    await queue.join()
    await asyncio.gather(*tasks, return_exceptions=True)



import argparse
import asyncio
import logging
import os
import sys

from .nrich import nrich
from .parse_internetdb import parse_internetdb

logger = logging.getLogger()
logging.basicConfig(level=os.getenv("LOGLEVEL", logging.INFO))


async def main(tasks: list):
    """main function to retrieve IP information from the shodan internetDB
    written in asyncio to speed up the requests.
    takes task list as argument to properly handle keyboard interrupts.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-input-file",
        "-i",
        type=argparse.FileType("r"),
        default=(None if sys.stdin.isatty() else sys.stdin),
        help="File to parse IPs from, if not given parse from stdin",
    )
    parser.add_argument(
        "-output-file", "-o", type=str, help="file to write IPs to", default=None
    )
    parser.add_argument(
        "-output-type",
        "-t",
        choices=["json", "str"],
        default="json",
        help="Format to output in",
    )
    parser.add_argument(
        "-skip_missing",
        "-s",
        action="store_false",
        default=True,
        help="skip IPs with no information on them.",
    )

    parser.add_argument("-mode", "-m", choices=["nrich", "parse"], default="nrich")

    args = parser.parse_args()

    if args.mode == "nrich":
        await nrich(args)
    elif args.mode == "parse":
        await parse_internetdb(args)


if __name__ == "__main__":
    tasks = []

    try:
        asyncio.run(main(tasks))
    except KeyboardInterrupt:
        if tasks:
            # Cancel our worker tasks.
            for task in tasks:
                task.cancel()



from dataclasses import dataclass, field
from ipaddress import IPv4Address, IPv6Address
from typing import Optional


@dataclass
class ShodanResult:
    """Dataclass that represents the returned result
    from the InternetDB"""

    ip: Optional[IPv4Address | IPv6Address] = None
    cpes: Optional[list[str]] = field(default_factory=list)
    hostnames: list[str] = field(default_factory=list)
    ports: list[int] = field(default_factory=list)
    tags: list[str] = field(default_factory=list)
    vulns: list[str] = field(default_factory=list)
    detail: str = ""
    empty: bool = field(init=False)

    def __post_init__(self):
        if self.detail == "No information available":
            self.empty = True

        self.empty = False
